
<div class="row">
<div class="col">
<button type="button" class="btn btn-sm btn-dark mb-2" id="{{ name }}-refresh"><i class="fas fa-sync-alt mr-2"></i>Refresh table manually</button>
</div>
</div>

<div class="row">
<div class="col small">

<table class="table table-striped table-hover" id="{{ name }}" width="100%">
<thead>
<tr>
<th>Session</th>
<th>Status</th>
<th>Condition</th>
<th>Last Page</th>
<th>Start Day</th>
<th>Start Time</th>
<th>Last Move</th>
<th>Group</th>
<th>Role</th>
</tr>

</thead>


</table>

</div>
</div>

<style>
#{{ name }}-refresh {float: right;}
#{{ name }} th, td {white-space: nowrap;}
</style>