
var render_table = function() {
    var table = $('#{{ name }}').DataTable(
        {"ajax": "{{ render_url }}", 
        "scrollX": true,
        "order": [[ 4, "desc" ]],
        "columns": [
            {% for col in cols %}
                {"data": "{{ col }}"},
            {% endfor %}
        ]}
    );
    
    return table
};

$(document).ready(function() {
    var table = render_table();
    
    // reload periodically
    //setInterval(function() {
    //    table.ajax.reload();
    //}, {{ refresh_interval }}*1000);
    
    // reload on click
    $("#{{ name }}-refresh").click(function() {
        table.ajax.reload();
    });
})